# This module contains functions from pabutools the were modified to generate random instances and profiles in a
# deterministic way. For documentation of the original functions, please refer to the pabutools documentation.

import random
from math import ceil

from pabutools.election import ApprovalProfile, Instance, Project, get_random_approval_profile


def get_random_election(
    num_projects: int = 3, min_cost: int = 1, max_cost: int = 4, num_agents: int = 5
) -> tuple[Instance, ApprovalProfile]:
    """Generates and returns a random election (Instance and ApprovalProfile) with the given parameters."""
    instance = get_random_instance(num_projects, min_cost, max_cost)
    profile = get_random_approval_profile(instance, num_agents)
    return instance, profile


def get_random_project(instance: Instance) -> Project:
    """Selects and returns a random Project from the given Instance."""
    return random.choice(sorted(instance))


def get_random_instance(num_projects: int, min_cost: int, max_cost: int) -> Instance:
    """
    Generates a random instance.

    Function adapted from pabutools, disallows not feasible projects.
    """
    inst = Instance()
    inst.update(
        Project(
            name=str(p),
            cost=random.randint(round(min_cost), round(max_cost)),
        )
        for p in range(round(num_projects))
    )
    inst.budget_limit = random.randint(ceil(max(p.cost for p in inst)), ceil(sum(p.cost for p in inst)))
    return inst



from pabumeasures._core import Comparator, Ordering, ProjectComparator
from pabumeasures.main import (
    Measure,
    greedy,
    greedy_measure,
    greedy_over_cost,
    greedy_over_cost_measure,
    mes_apr,
    mes_apr_measure,
    mes_cost,
    mes_cost_measure,
    phragmen,
    phragmen_measure,
)

__all__ = [
    "Measure",
    "Comparator",
    "Ordering",
    "ProjectComparator",
    "greedy",
    "greedy_measure",
    "greedy_over_cost",
    "greedy_over_cost_measure",
    "mes_apr",
    "mes_apr_measure",
    "mes_cost",
    "mes_cost_measure",
    "phragmen",
    "phragmen_measure",
]



import pytest
from pabutools.election import ApprovalBallot, ApprovalProfile, Instance, Project

import pabumeasures


def test_error_on_same_names():
    p1 = Project("p", 2)
    p2 = Project("p", 1)
    instance = Instance([p1, p2], 2)
    if len(instance) != 2:
        # pabutools removes duplicates
        assert len(instance) == 1
    else:
        # if pabutools does not remove duplicates (for some reason), we should raise an error
        profile = ApprovalProfile(
            [
                ApprovalBallot([p1]),
                ApprovalBallot([p2]),
                ApprovalBallot([p1, p2]),
            ]
        )

        with pytest.raises(ValueError, match=r"names .+ unique"):
            pabumeasures.greedy(instance, profile)


def test_error_on_negative_cost():
    p1 = Project("p1", -2)
    p2 = Project("p2", 1)
    instance = Instance([p1, p2], 2)
    profile = ApprovalProfile(
        [
            ApprovalBallot([p1]),
            ApprovalBallot([p2]),
            ApprovalBallot([p1, p2]),
        ]
    )

    with pytest.raises(ValueError, match=r"costs .+ positive"):
        pabumeasures.greedy(instance, profile)


def test_error_on_cost_larger_than_budget():
    p1 = Project("p1", 11)
    p2 = Project("p2", 1)
    instance = Instance([p1, p2], 10)
    profile = ApprovalProfile(
        [
            ApprovalBallot([p1]),
            ApprovalBallot([p2]),
        ]
    )

    with pytest.raises(ValueError, match=r"costs .+ exceed"):
        pabumeasures.greedy(instance, profile)


def test_error_on_huge_budget():
    p1 = Project("p1", 2)
    p2 = Project("p2", 1)
    instance = Instance([p1, p2], int(2e9))
    profile = ApprovalProfile(
        [
            ApprovalBallot([p1]),
            ApprovalBallot([p2]),
        ]
    )

    with pytest.raises(ValueError, match=r"[Bb]udget limit .+ exceed"):
        pabumeasures.greedy(instance, profile)



import random
import string

import pytest

from pabumeasures._core import Comparator, Ordering, ProjectComparator, ProjectEmbedding


def random_project_embedding(
    min_cost: int, max_cost: int, name_length: int = 5, max_approvers: int = 3
) -> ProjectEmbedding:
    cost = random.randint(min_cost, max_cost)
    name = "".join(random.choices(string.ascii_letters, k=name_length))
    approvers = random.sample(range(max_approvers), random.randint(0, max_approvers))
    return ProjectEmbedding(cost, name, approvers)


random.seed(42)
projects = [random_project_embedding(1, 3) for _ in range(200)]

test_cases = [
    (lambda p: (p.cost, p.name), ProjectComparator.ByCostAsc, "ByCostAsc"),
    (lambda p: (p.cost, p.name), ProjectComparator(Comparator.COST, Ordering.ASCENDING), "ByCostAsc_explicit"),
    (lambda p: (-p.cost, p.name), ProjectComparator.ByCostDesc, "ByCostDesc"),
    (lambda p: (-p.cost, p.name), ProjectComparator(Comparator.COST, Ordering.DESCENDING), "ByCostDesc_explicit"),
    (lambda p: (p.name), ProjectComparator.ByNameAsc, "ByNameAsc"),
    (lambda p: (p.name), ProjectComparator([]), "ByNameAsc_implicit"),
    (lambda p: (p.name), ProjectComparator(Comparator.LEXICOGRAPHIC, Ordering.ASCENDING), "ByNameAsc_explicit"),
    (lambda p: "".join(chr(255 - ord(c)) for c in p.name), ProjectComparator.ByNameDesc, "ByNameDesc"),
    (
        lambda p: "".join(chr(255 - ord(c)) for c in p.name),
        ProjectComparator(Comparator.LEXICOGRAPHIC, Ordering.DESCENDING),
        "ByNameDesc_explicit",
    ),
    (
        lambda p: (len(p.approvers), p.name),
        ProjectComparator(Comparator.VOTES, Ordering.ASCENDING),
        "ByVotesAsc_explicit",
    ),
    (lambda p: (-len(p.approvers), p.name), ProjectComparator.ByVotesDesc, "ByVotesDesc"),
    (
        lambda p: (-len(p.approvers), p.name),
        ProjectComparator(Comparator.VOTES, Ordering.DESCENDING),
        "ByVotesDesc_explicit",
    ),
    (lambda p: (p.cost, -len(p.approvers), p.name), ProjectComparator.ByCostAscThenVotesDesc, "ByCostAscThenVotesDesc"),
    (
        lambda p: (p.cost, -len(p.approvers), p.name),
        ProjectComparator([(Comparator.COST, Ordering.ASCENDING), (Comparator.VOTES, Ordering.DESCENDING)]),
        "ByCostAscThenVotesDesc_explicit",
    ),
    (
        lambda p: (-p.cost, -len(p.approvers), p.name),
        ProjectComparator.ByCostDescThenVotesDesc,
        "ByCostDescThenVotesDesc",
    ),
    (
        lambda p: (-p.cost, -len(p.approvers), p.name),
        ProjectComparator([(Comparator.COST, Ordering.DESCENDING), (Comparator.VOTES, Ordering.DESCENDING)]),
        "ByCostDescThenVotesDesc_explicit",
    ),
    (
        lambda p: (p.cost, -len(p.approvers), p.name),
        ProjectComparator(
            [
                (Comparator.COST, Ordering.ASCENDING),
                (Comparator.VOTES, Ordering.DESCENDING),
                (Comparator.LEXICOGRAPHIC, Ordering.ASCENDING),
            ]
        ),
        "ByCostVotesLex",
    ),
]


@pytest.mark.parametrize("key_func,comparator,name", test_cases)
def test_project_comparator_basic(key_func, comparator, name):
    sorted_projects = list(projects)
    sorted_projects.sort(key=key_func)

    for i in range(len(sorted_projects)):
        for j in range(i + 1, len(sorted_projects)):
            same_key = key_func(sorted_projects[i]) == key_func(sorted_projects[j])

            if same_key:
                assert not comparator(sorted_projects[i], sorted_projects[j])
                assert not comparator(sorted_projects[j], sorted_projects[i])
            else:
                assert comparator(sorted_projects[i], sorted_projects[j]) != comparator(
                    sorted_projects[j], sorted_projects[i]
                )
                assert comparator(sorted_projects[i], sorted_projects[j])



import random
from itertools import chain, combinations

import pytest
from pabutools.election import ApprovalBallot
from utils import get_random_election, get_random_project

import pabumeasures
from pabumeasures import Measure


def _powerset(iterable):
    # powerset([1,2,3]) → () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))


NUMBER_OF_TIMES = 500


@pytest.mark.parametrize("seed", list(range(NUMBER_OF_TIMES)))
@pytest.mark.parametrize(
    "rule,rule_measure",
    [
        (pabumeasures.greedy, pabumeasures.greedy_measure),
        (pabumeasures.greedy_over_cost, pabumeasures.greedy_over_cost_measure),
        (pabumeasures.mes_apr, pabumeasures.mes_apr_measure),
        (pabumeasures.mes_cost, pabumeasures.mes_cost_measure),
        (pabumeasures.phragmen, pabumeasures.phragmen_measure),
    ],
    ids=["greedy", "greedy_over_cost", "mes_apr", "mes_cost", "phragmen"],
)
def test_optimist_add_measure(seed, rule, rule_measure):
    random.seed(seed)
    instance, profile = get_random_election()
    project = get_random_project(instance)
    allocation = rule(instance, profile)
    result = rule_measure(instance, profile, project, Measure.ADD_APPROVAL_OPTIMIST)
    if project in allocation:
        assert result == 0
    else:
        non_approvers = [ballot for ballot in profile if project not in ballot]
        if result is None:
            for na in non_approvers:
                na.add(project)
            assert project not in rule(instance, profile), "result should not be None"
        else:
            assert 1 <= result <= len(non_approvers)
            for new_approvers in _powerset(non_approvers):
                for na in new_approvers:
                    na.add(project)
                if project in rule(instance, profile):
                    # here we use the fact that subsets are generated in increasing order of size
                    assert result == len(new_approvers)
                    return
                for na in new_approvers:
                    na.remove(project)


@pytest.mark.parametrize("seed", list(range(NUMBER_OF_TIMES)))
@pytest.mark.parametrize(
    "rule,rule_measure",
    [
        (pabumeasures.greedy, pabumeasures.greedy_measure),
        (pabumeasures.greedy_over_cost, pabumeasures.greedy_over_cost_measure),
        (pabumeasures.mes_apr, pabumeasures.mes_apr_measure),
        (pabumeasures.mes_cost, pabumeasures.mes_cost_measure),
        (pabumeasures.phragmen, pabumeasures.phragmen_measure),
    ],
    ids=["greedy", "greedy_over_cost", "mes_apr", "mes_cost", "phragmen"],
)
def test_pessimist_add_measure(seed, rule, rule_measure):
    random.seed(seed)
    instance, profile = get_random_election()
    project = get_random_project(instance)
    allocation = rule(instance, profile)
    result = rule_measure(instance, profile, project, Measure.ADD_APPROVAL_PESSIMIST)
    if project in allocation:
        assert result == 0
    else:
        non_approvers = [ballot for ballot in profile if project not in ballot]
        for expected_result in range(1, len(non_approvers) + 1):
            ok = True
            for new_approvers in combinations(non_approvers, expected_result):
                for na in new_approvers:
                    na.add(project)
                if project not in rule(instance, profile):
                    ok = False
                for na in new_approvers:
                    na.remove(project)
            if ok:
                assert result == expected_result
                return
        assert result is None


@pytest.mark.parametrize("seed", list(range(NUMBER_OF_TIMES)))
@pytest.mark.parametrize(
    "rule,rule_measure",
    [
        (pabumeasures.greedy, pabumeasures.greedy_measure),
        (pabumeasures.greedy_over_cost, pabumeasures.greedy_over_cost_measure),
        (pabumeasures.mes_apr, pabumeasures.mes_apr_measure),
        (pabumeasures.mes_cost, pabumeasures.mes_cost_measure),
        (pabumeasures.phragmen, pabumeasures.phragmen_measure),
    ],
    ids=["greedy", "greedy_over_cost", "mes_apr", "mes_cost", "phragmen"],
)
def test_singleton_add_measure(seed, rule, rule_measure):
    random.seed(seed)
    instance, profile = get_random_election()
    project = get_random_project(instance)
    allocation = rule(instance, profile)
    result = rule_measure(instance, profile, project, Measure.ADD_SINGLETON)

    if rule in [pabumeasures.mes_apr, pabumeasures.mes_cost] and result is None:
        assert instance.budget_limit == project.cost
    else:
        assert result is not None

        if project in allocation:
            assert result == 0
        else:
            assert result >= 1
            for i in range(len(profile), len(profile) + result):
                profile.append(ApprovalBallot({project}, name=f"SingletonAppBallot {i}"))
            assert project in rule(instance, profile)
            profile.pop()
            assert project not in rule(instance, profile)


@pytest.mark.parametrize("seed", list(range(NUMBER_OF_TIMES)))
@pytest.mark.parametrize(
    "rule,rule_measure",
    [
        (pabumeasures.greedy, pabumeasures.greedy_measure),
        (pabumeasures.greedy_over_cost, pabumeasures.greedy_over_cost_measure),
        (pabumeasures.mes_apr, pabumeasures.mes_apr_measure),
        (pabumeasures.mes_cost, pabumeasures.mes_cost_measure),
        (pabumeasures.phragmen, pabumeasures.phragmen_measure),
    ],
    ids=["greedy", "greedy_over_cost", "mes_apr", "mes_cost", "phragmen"],
)
def test_cost_reduction_measure(seed, rule, rule_measure):
    random.seed(seed)
    instance, profile = get_random_election()
    project = get_random_project(instance)
    allocation = rule(instance, profile)
    result = rule_measure(instance, profile, project, Measure.COST_REDUCTION)

    if project in allocation:
        assert result == project.cost
    else:
        if result > 0:
            project.cost = result
            assert project in rule(instance, profile)
        else:
            assert result == 0

        project.cost = result + 1
        assert project not in rule(instance, profile)



from pabutools.election import ApprovalBallot, ApprovalProfile, Instance, Project

from pabumeasures import Measure, mes_cost, mes_cost_measure


def test_example():
    p1 = Project("p1", 1)
    p2 = Project("p2", 1)
    p3 = Project("p3", 3)

    b1 = ApprovalBallot([p1, p2])
    b2 = ApprovalBallot([p1, p2, p3])
    b3 = ApprovalBallot([p3])

    instance = Instance([p1, p2, p3], budget_limit=3)
    profile = ApprovalProfile([b1, b2, b3])

    assert mes_cost(instance, profile) == [p1, p2]
    assert mes_cost_measure(instance, profile, p3, Measure.ADD_APPROVAL_OPTIMIST) == 1



import enum
from typing import overload

# ========== project enums ==========

class Comparator(enum.Enum):
    COST: Comparator
    VOTES: Comparator
    LEXICOGRAPHIC: Comparator

class Ordering(enum.Enum):
    ASCENDING: Ordering
    DESCENDING: Ordering

# ========== project classes ==========

class Election:
    def __init__(self, budget: int, num_of_voters: int, projects: list[ProjectEmbedding]) -> None: ...
    @property
    def budget(self) -> int: ...
    @property
    def num_of_voters(self) -> int: ...
    @property
    def projects(self) -> list[ProjectEmbedding]: ...

class ProjectEmbedding:
    def __init__(self, cost: int, name: str, approvers: list[int] = ...) -> None: ...
    @property
    def cost(self) -> int: ...
    @property
    def name(self) -> str: ...
    @property
    def approvers(self) -> list[int]: ...
    @property
    def num_of_approvers(self) -> int: ...

class ProjectComparator:
    ByCostAsc: ProjectComparator
    ByCostDesc: ProjectComparator
    ByNameAsc: ProjectComparator
    ByNameDesc: ProjectComparator
    ByVotesDesc: ProjectComparator
    ByCostAscThenVotesDesc: ProjectComparator
    ByCostDescThenVotesDesc: ProjectComparator

    @overload
    def __init__(self, criteria: list[tuple[Comparator, Ordering]]) -> None: ...
    @overload
    def __init__(self, comparator: Comparator, ordering: Ordering) -> None: ...
    def __init__(self, *args, **kwargs) -> None: ...
    def __call__(self, lhs: ProjectEmbedding, rhs: ProjectEmbedding) -> bool: ...

# ========== rules ==========

def greedy(election: Election, tie_breaking: ProjectComparator) -> list[ProjectEmbedding]: ...
def greedy_over_cost(election: Election, tie_breaking: ProjectComparator) -> list[ProjectEmbedding]: ...
def mes_apr(election: Election, tie_breaking: ProjectComparator) -> list[ProjectEmbedding]: ...
def mes_cost(election: Election, tie_breaking: ProjectComparator) -> list[ProjectEmbedding]: ...
def phragmen(election: Election, tie_breaking: ProjectComparator) -> list[ProjectEmbedding]: ...

# ========== optimist-add ==========

def optimist_add_for_greedy(election: Election, p: int, tie_breaking: ProjectComparator) -> int | None: ...
def optimist_add_for_greedy_over_cost(election: Election, p: int, tie_breaking: ProjectComparator) -> int | None: ...
def optimist_add_for_mes_apr(election: Election, p: int, tie_breaking: ProjectComparator) -> int | None: ...
def optimist_add_for_mes_cost(election: Election, p: int, tie_breaking: ProjectComparator) -> int | None: ...
def optimist_add_for_phragmen(election: Election, p: int, tie_breaking: ProjectComparator) -> int | None: ...

# ========== pessimist-add ==========

def pessimist_add_for_greedy(election: Election, p: int, tie_breaking: ProjectComparator) -> int | None: ...
def pessimist_add_for_greedy_over_cost(election: Election, p: int, tie_breaking: ProjectComparator) -> int | None: ...
def pessimist_add_for_mes_apr(election: Election, p: int, tie_breaking: ProjectComparator) -> int | None: ...
def pessimist_add_for_mes_cost(election: Election, p: int, tie_breaking: ProjectComparator) -> int | None: ...
def pessimist_add_for_phragmen(election: Election, p: int, tie_breaking: ProjectComparator) -> int | None: ...

# ========== singleton-add ==========

def singleton_add_for_greedy(election: Election, p: int, tie_breaking: ProjectComparator) -> int | None: ...
def singleton_add_for_greedy_over_cost(election: Election, p: int, tie_breaking: ProjectComparator) -> int | None: ...
def singleton_add_for_mes_apr(election: Election, p: int, tie_breaking: ProjectComparator) -> int | None: ...
def singleton_add_for_mes_cost(election: Election, p: int, tie_breaking: ProjectComparator) -> int | None: ...
def singleton_add_for_phragmen(election: Election, p: int, tie_breaking: ProjectComparator) -> int | None: ...

# ========== cost-reduction ==========

def cost_reduction_for_greedy(election: Election, p: int, tie_breaking: ProjectComparator) -> int: ...
def cost_reduction_for_greedy_over_cost(election: Election, p: int, tie_breaking: ProjectComparator) -> int: ...
def cost_reduction_for_mes_apr(election: Election, p: int, tie_breaking: ProjectComparator) -> int: ...
def cost_reduction_for_mes_cost(election: Election, p: int, tie_breaking: ProjectComparator) -> int: ...
def cost_reduction_for_phragmen(election: Election, p: int, tie_breaking: ProjectComparator) -> int: ...



from enum import Enum, auto

from pabutools.election.ballot import FrozenBallot
from pabutools.election.instance import Instance, Project
from pabutools.election.profile import ApprovalProfile, Profile
from pabutools.rules import BudgetAllocation

from pabumeasures import ProjectComparator, _core


class Measure(Enum):
    COST_REDUCTION = auto()
    ADD_APPROVAL_OPTIMIST = auto()
    ADD_APPROVAL_PESSIMIST = auto()
    ADD_SINGLETON = auto()


def _translate_input_format(instance: Instance, profile: Profile) -> tuple[_core.Election, dict[str, Project]]:
    if not isinstance(instance, Instance):
        raise TypeError("Instance must be of type Instance")
    if not isinstance(profile, ApprovalProfile):
        raise TypeError("Profile must be of type ApprovalProfile")
    if len([project.name for project in instance]) != len({project.name for project in instance}):
        raise ValueError("Project names must be unique in the instance")
    if any(project.cost <= 0 for project in instance):
        raise ValueError("Project costs must be positive")
    if any(project.cost > instance.budget_limit for project in instance):
        raise ValueError("Project costs must not exceed the budget limit")
    if instance.budget_limit > 1_000_000_000:
        raise ValueError("Budget limit must not exceed 1 billion")

    projects: list[Project] = sorted(instance)
    frozen_ballots: list[tuple[int, FrozenBallot]] = [(i, ballot.frozen()) for i, ballot in enumerate(profile)]
    total_budget = int(instance.budget_limit)
    approvers: dict[str, list[int]] = {project.name: [] for project in projects}
    for i, frozen_ballot in frozen_ballots:
        for project in frozen_ballot:
            approvers[project.name].append(i)
    project_embeddings: list[_core.ProjectEmbedding] = [
        _core.ProjectEmbedding(int(project.cost), project.name, approvers[project.name]) for project in projects
    ]
    name_to_project: dict[str, Project] = {project.name: project for project in projects}
    return _core.Election(total_budget, len(profile), project_embeddings), name_to_project


def greedy(
    instance: Instance, profile: Profile, tie_breaking: ProjectComparator = ProjectComparator.ByCostAsc
) -> BudgetAllocation:
    election, name_to_project = _translate_input_format(instance, profile)
    result = _core.greedy(election, tie_breaking)
    return BudgetAllocation(name_to_project[project_embeding.name] for project_embeding in result)


def greedy_measure(
    instance: Instance,
    profile: Profile,
    project: Project,
    measure: Measure,
    tie_breaking: ProjectComparator = ProjectComparator.ByCostAsc,
) -> int | None:
    election, _ = _translate_input_format(instance, profile)
    p = sorted(instance).index(project)
    match measure:
        case Measure.COST_REDUCTION:
            return _core.cost_reduction_for_greedy(election, p, tie_breaking)
        case Measure.ADD_APPROVAL_OPTIMIST:
            return _core.optimist_add_for_greedy(election, p, tie_breaking)
        case Measure.ADD_APPROVAL_PESSIMIST:
            return _core.pessimist_add_for_greedy(election, p, tie_breaking)
        case Measure.ADD_SINGLETON:
            return _core.singleton_add_for_greedy(election, p, tie_breaking)


def greedy_over_cost(
    instance: Instance, profile: Profile, tie_breaking: ProjectComparator = ProjectComparator.ByCostAsc
) -> BudgetAllocation:
    election, name_to_project = _translate_input_format(instance, profile)
    result = _core.greedy_over_cost(election, tie_breaking)
    return BudgetAllocation(name_to_project[project_embeding.name] for project_embeding in result)


def greedy_over_cost_measure(
    instance: Instance,
    profile: Profile,
    project: Project,
    measure: Measure,
    tie_breaking: ProjectComparator = ProjectComparator.ByCostAsc,
) -> int | None:
    election, _ = _translate_input_format(instance, profile)
    p = sorted(instance).index(project)
    match measure:
        case Measure.COST_REDUCTION:
            return _core.cost_reduction_for_greedy_over_cost(election, p, tie_breaking)
        case Measure.ADD_APPROVAL_OPTIMIST:
            return _core.optimist_add_for_greedy_over_cost(election, p, tie_breaking)
        case Measure.ADD_APPROVAL_PESSIMIST:
            return _core.pessimist_add_for_greedy_over_cost(election, p, tie_breaking)
        case Measure.ADD_SINGLETON:
            return _core.singleton_add_for_greedy_over_cost(election, p, tie_breaking)


def mes_apr(
    instance: Instance, profile: Profile, tie_breaking: ProjectComparator = ProjectComparator.ByCostAsc
) -> BudgetAllocation:
    election, name_to_project = _translate_input_format(instance, profile)
    result = _core.mes_apr(election, tie_breaking)
    return BudgetAllocation(name_to_project[project_embeding.name] for project_embeding in result)


def mes_apr_measure(
    instance: Instance,
    profile: Profile,
    project: Project,
    measure: Measure,
    tie_breaking: ProjectComparator = ProjectComparator.ByCostAsc,
) -> int | None:
    election, _ = _translate_input_format(instance, profile)
    p = sorted(instance).index(project)
    match measure:
        case Measure.COST_REDUCTION:
            return _core.cost_reduction_for_mes_apr(election, p, tie_breaking)
        case Measure.ADD_APPROVAL_OPTIMIST:
            return _core.optimist_add_for_mes_apr(election, p, tie_breaking)
        case Measure.ADD_APPROVAL_PESSIMIST:
            return _core.pessimist_add_for_mes_apr(election, p, tie_breaking)
        case Measure.ADD_SINGLETON:
            return _core.singleton_add_for_mes_apr(election, p, tie_breaking)


def mes_cost(
    instance: Instance, profile: Profile, tie_breaking: ProjectComparator = ProjectComparator.ByCostAsc
) -> BudgetAllocation:
    election, name_to_project = _translate_input_format(instance, profile)
    result = _core.mes_cost(election, tie_breaking)
    return BudgetAllocation(name_to_project[project_embeding.name] for project_embeding in result)


def mes_cost_measure(
    instance: Instance,
    profile: Profile,
    project: Project,
    measure: Measure,
    tie_breaking: ProjectComparator = ProjectComparator.ByCostAsc,
) -> int | None:
    election, _ = _translate_input_format(instance, profile)
    p = sorted(instance).index(project)
    match measure:
        case Measure.COST_REDUCTION:
            return _core.cost_reduction_for_mes_cost(election, p, tie_breaking)
        case Measure.ADD_APPROVAL_OPTIMIST:
            return _core.optimist_add_for_mes_cost(election, p, tie_breaking)
        case Measure.ADD_APPROVAL_PESSIMIST:
            return _core.pessimist_add_for_mes_cost(election, p, tie_breaking)
        case Measure.ADD_SINGLETON:
            return _core.singleton_add_for_mes_cost(election, p, tie_breaking)


def phragmen(
    instance: Instance, profile: Profile, tie_breaking: ProjectComparator = ProjectComparator.ByCostAsc
) -> BudgetAllocation:
    election, name_to_project = _translate_input_format(instance, profile)
    result = _core.phragmen(election, tie_breaking)
    return BudgetAllocation(name_to_project[project_embeding.name] for project_embeding in result)


def phragmen_measure(
    instance: Instance,
    profile: Profile,
    project: Project,
    measure: Measure,
    tie_breaking: ProjectComparator = ProjectComparator.ByCostAsc,
) -> int | None:
    election, _ = _translate_input_format(instance, profile)
    p = sorted(instance).index(project)
    match measure:
        case Measure.COST_REDUCTION:
            return _core.cost_reduction_for_phragmen(election, p, tie_breaking)
        case Measure.ADD_APPROVAL_OPTIMIST:
            return _core.optimist_add_for_phragmen(election, p, tie_breaking)
        case Measure.ADD_APPROVAL_PESSIMIST:
            return _core.pessimist_add_for_phragmen(election, p, tie_breaking)
        case Measure.ADD_SINGLETON:
            return _core.singleton_add_for_phragmen(election, p, tie_breaking)



import glob
import random

import pytest
from pabutools.election import ApprovalProfile, Cardinality_Sat, Cost_Sat, parse_pabulib
from pabutools.rules import greedy_utilitarian_welfare, method_of_equal_shares, sequential_phragmen
from pabutools.tiebreaking import TieBreakingRule
from utils import get_random_election

import pabumeasures

# redefined pabutools tie-breaking rules to include project name as a secondary criterion
min_cost_tie_breaking = TieBreakingRule(lambda inst, prof, proj: (proj.cost, proj.name))
max_cost_tie_breaking = TieBreakingRule(lambda inst, prof, proj: (-proj.cost, proj.name))

test_files = glob.glob("./data/*.pb")
NUMBER_OF_TIMES = 500


@pytest.mark.parametrize("file", test_files)
def test_greedy(file):
    instance, profile = parse_pabulib(file)
    pabutools_result = greedy_utilitarian_welfare(
        instance, profile, sat_class=Cost_Sat, tie_breaking=min_cost_tie_breaking
    )
    result = pabumeasures.greedy(instance, profile)

    assert sorted(pabutools_result) == sorted(result)


@pytest.mark.parametrize("seed", list(range(NUMBER_OF_TIMES)))
def test_greedy_random(seed):
    random.seed(seed)
    instance, profile = get_random_election()
    pabutools_result = greedy_utilitarian_welfare(
        instance, profile, sat_class=Cost_Sat, tie_breaking=min_cost_tie_breaking
    )
    result = pabumeasures.greedy(instance, profile)

    assert sorted(pabutools_result) == sorted(result)


@pytest.mark.parametrize("seed", list(range(NUMBER_OF_TIMES)))
def test_greedy_random_different_comparator(seed):
    random.seed(seed)
    instance, profile = get_random_election()
    pabutools_result = greedy_utilitarian_welfare(
        instance, profile, sat_class=Cost_Sat, tie_breaking=max_cost_tie_breaking
    )
    result = pabumeasures.greedy(
        instance,
        profile,
        pabumeasures.ProjectComparator(pabumeasures.Comparator.COST, pabumeasures.Ordering.DESCENDING),
    )

    assert sorted(pabutools_result) == sorted(result)


@pytest.mark.parametrize("file", test_files)
def test_greedy_over_cost(file):
    instance, profile = parse_pabulib(file)
    pabutools_result = greedy_utilitarian_welfare(
        instance, profile, sat_class=Cardinality_Sat, tie_breaking=min_cost_tie_breaking
    )
    result = pabumeasures.greedy_over_cost(instance, profile)

    assert sorted(pabutools_result) == sorted(result)


@pytest.mark.parametrize("seed", list(range(NUMBER_OF_TIMES)))
def test_greedy_over_cost_random(seed):
    random.seed(seed)
    instance, profile = get_random_election()
    pabutools_result = greedy_utilitarian_welfare(
        instance, profile, sat_class=Cardinality_Sat, tie_breaking=min_cost_tie_breaking
    )
    result = pabumeasures.greedy_over_cost(instance, profile)

    assert sorted(pabutools_result) == sorted(result)


@pytest.mark.parametrize("seed", list(range(NUMBER_OF_TIMES)))
def test_greedy_over_cost_random_different_comparator(seed):
    random.seed(seed)
    instance, profile = get_random_election()
    pabutools_result = greedy_utilitarian_welfare(
        instance, profile, sat_class=Cardinality_Sat, tie_breaking=max_cost_tie_breaking
    )
    result = pabumeasures.greedy_over_cost(
        instance,
        profile,
        pabumeasures.ProjectComparator(pabumeasures.Comparator.COST, pabumeasures.Ordering.DESCENDING),
    )

    assert sorted(pabutools_result) == sorted(result)


@pytest.mark.parametrize("file", test_files)
def test_mes_apr(file):
    instance, profile = parse_pabulib(file)
    pabutools_result = method_of_equal_shares(
        instance, profile, sat_class=Cardinality_Sat, tie_breaking=min_cost_tie_breaking
    )
    result = pabumeasures.mes_apr(instance, profile)

    assert sorted(pabutools_result) == sorted(result)


@pytest.mark.parametrize("seed", list(range(NUMBER_OF_TIMES)))
def test_mes_apr_random(seed):
    random.seed(seed)
    instance, profile = get_random_election()
    pabutools_result = method_of_equal_shares(
        instance, profile, sat_class=Cardinality_Sat, tie_breaking=min_cost_tie_breaking
    )
    result = pabumeasures.mes_apr(instance, profile)

    assert sorted(pabutools_result) == sorted(result)


@pytest.mark.parametrize("file", test_files)
def test_mes_cost(file):
    instance, profile = parse_pabulib(file)
    pabutools_result = method_of_equal_shares(instance, profile, sat_class=Cost_Sat, tie_breaking=min_cost_tie_breaking)
    result = pabumeasures.mes_cost(instance, profile)

    assert sorted(pabutools_result) == sorted(result)


@pytest.mark.parametrize("seed", list(range(NUMBER_OF_TIMES)))
def test_mes_cost_random(seed):
    random.seed(seed)
    instance, profile = get_random_election()
    pabutools_result = method_of_equal_shares(instance, profile, sat_class=Cost_Sat, tie_breaking=min_cost_tie_breaking)
    result = pabumeasures.mes_cost(instance, profile)

    assert sorted(pabutools_result) == sorted(result)


@pytest.mark.parametrize("file", test_files)
def test_phragmen(file):
    instance, profile = parse_pabulib(file)
    assert isinstance(profile, ApprovalProfile)  # for type checking
    pabutools_result = sequential_phragmen(instance, profile, tie_breaking=min_cost_tie_breaking)
    result = pabumeasures.phragmen(instance, profile)

    assert sorted(pabutools_result) == sorted(result)


@pytest.mark.parametrize("seed", list(range(NUMBER_OF_TIMES)))
def test_phragmen_random(seed):
    random.seed(seed)
    instance, profile = get_random_election()
    pabutools_result = sequential_phragmen(instance, profile, tie_breaking=min_cost_tie_breaking)
    result = pabumeasures.phragmen(instance, profile)

    assert sorted(pabutools_result) == sorted(result)
